#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "m_pd.h"
#include "g_canvas.h"
#include "g_all_guis.h"

static t_class *ext_class;

#define DEBUG 1

#ifdef DEBUG

#define logd(fmt, ...) \
    printf(fmt, ##__VA_ARGS__); \
    fflush(stdout)
#else

#define logd(fmt, ...)

#endif

#define sys_vguid(fmt, ...) \
    logd(fmt, ##__VA_ARGS__); \
    sys_vgui(fmt, ##__VA_ARGS__)

typedef struct _ext {
    t_iemgui x_gui;
} t_ext;

t_widgetbehavior ext_widgetbehavior;

void get_rect(t_ext *x, t_glist *glist, int *xpos, int *ypos, int *fxpos, int *fypos, int *lxpos, int *lypos) {
    *xpos = text_xpix(&x->x_gui.x_obj, glist);
    *ypos = text_ypix(&x->x_gui.x_obj, glist);
    *fxpos = *xpos + x->x_gui.x_w;
    *fypos = *ypos + x->x_gui.x_h;
    *lxpos = *xpos + x->x_gui.x_ldx;
    *lypos = *ypos + x->x_gui.x_ldy;
}

char *get_name(t_ext *x, t_glist *glist) {    
    static char buffer[100]; // not thread-safe
    t_canvas *canvas = glist_getcanvas(glist);
    sprintf(buffer, ".x%lx.c", x);
    return buffer;
}

char *get_col(int col) {
    static char buffer[100]; // not thread-safe
    sprintf(buffer, "#%6.6x", col);
    return buffer;
}

char *get_bcol(t_ext *x) {
    return get_col(x->x_gui.x_bcol);
}

char *get_lcol(t_ext *x) {
    return x->x_gui.x_fsf.x_selected ? get_col(IEM_GUI_COLOR_SELECTED) : get_lcol(x);
}

char *get_scol(t_ext *x) {
    return x->x_gui.x_fsf.x_selected ? get_col(IEM_GUI_COLOR_SELECTED) : get_col(IEM_GUI_COLOR_NORMAL);
}

char *get_font(t_ext *x, char *mod) {
    static char buffer[100]; // not thread-safe
    sprintf(buffer, "{%s %d %s}", x->x_gui.x_font, x->x_gui.x_fontsize, mod);
    return buffer;
}

char *get_lname(t_ext *x) {
    return strcmp(x->x_gui.x_lab->s_name, "empty") ? x->x_gui.x_lab->s_name : "";
}

void ext_draw_new(t_ext *x, t_glist *glist) {
    logd("ext_draw_new\n");
    int xpos, ypos, fxpos, fypos, lxpos, lypos;
    get_rect(x, glist, &xpos, &ypos, &fxpos, &fypos, &lxpos, &lypos);
    char *name = get_name(x, glist);
    sys_vguid("%s create rectangle %d %d %d %d -fill %s -tags %s.base\n",
             name, xpos, ypos, fxpos, fypos, get_bcol(x), name);
    sys_vguid("%s create text %d %d -text {%s} -anchor w -font %s -fill %s -tags %s.label\n",
             name, lxpos, lypos, get_lname(x), get_font(x, "bold"), get_lcol(x), name);
}

void ext_draw_move(t_ext *x, t_glist *glist) {
    logd("ext_draw_move\n");
    int xpos, ypos, fxpos, fypos, lxpos, lypos;
    get_rect(x, glist, &xpos, &ypos, &fxpos, &fypos, &lxpos, &lypos);
    char *name = get_name(x, glist);
    sys_vguid("%s coords %s.base %d %d %d %d\n", name, name, xpos, ypos, fxpos, fypos);
    sys_vguid("%s coords %%s.label %d %d\n", name, name, lxpos, lypos);
}

void ext_draw_erase(t_ext *x, t_glist *glist) {
    logd("ext_draw_erase\n");
    char *name = get_name(x, glist);
    sys_vguid("%s delete %s.base\n", name, name);
    sys_vguid("%s delete %s.label\n", name, name);
}

void ext_draw_config(t_ext *x, t_glist *glist) {
    logd("ext_draw_config\n");
    char *name = get_name(x, glist);
    sys_vguid("%s itemconfigure %s.base -fill %s\n", name, name, get_bcol(x));
    sys_vguid("%s itemconfigure %s.label -font %s -fill %s -text {%s} \n",
             name, name, get_font(x, "bold"), get_lcol(x), get_lname(x));
}

void ext_draw_select(t_ext *x, t_glist *glist) {
    logd("ext_draw_select\n");
    char *name = get_name(x, glist);
    sys_vguid(".%s itemconfigure %s.base -outline %s\n", name, name, get_scol(x));
    sys_vguid(".%s itemconfigure %s.label -fill %s\n", name, name, get_lcol(x));
}

void ext_draw(t_ext *x, t_glist *glist, int mode) {
    logd("ext_draw\n");
    switch (mode) {
    case IEM_GUI_DRAW_MODE_NEW: ext_draw_new(x, glist); break;
    case IEM_GUI_DRAW_MODE_MOVE: ext_draw_move(x, glist); break;
    case IEM_GUI_DRAW_MODE_SELECT: ext_draw_select(x, glist); break;
    case IEM_GUI_DRAW_MODE_ERASE: ext_draw_erase(x, glist); break;
    case IEM_GUI_DRAW_MODE_CONFIG: ext_draw_config(x, glist); break;
    }
}

static void ext_getrect(t_gobj *z, t_glist *glist, int *xpos, int *ypos, int *fxpos, int *fypos)
{
    logd("ext_getrect\n");
    t_ext *x = (t_ext *) z;
    int lxpos, lypos;
    get_rect(x, glist, xpos, ypos, fxpos, fypos, &lxpos, &lypos);
}

void *ext_new(t_symbol *s, int argc, t_atom *argv)
{
    t_ext *x = (t_ext *) pd_new(ext_class);
    
    int bflcol[] = { -262144, -1, -1 };
    int w = 100, h = 60;
    int ldx = 17, ldy = 7;
    int fs = 10;
    iem_inttosymargs(&x->x_gui.x_isa, 0);
    iem_inttofstyle(&x->x_gui.x_fsf, 0);

    if((argc == 13)
       && IS_A_FLOAT(argv, 0) // width
       && IS_A_FLOAT(argv, 1) // height
       && IS_A_FLOAT(argv, 2) // loadinit = n, scale = n >> 20, flashed = locked = reverse=0
       && (IS_A_SYMBOL(argv, 3) || IS_A_FLOAT(argv, 3)) // x_snd name
       && (IS_A_SYMBOL(argv, 4) || IS_A_FLOAT(argv, 4)) // x_rcv name
       && (IS_A_SYMBOL(argv, 5) || IS_A_FLOAT(argv, 5)) // x_lab name
       && IS_A_FLOAT(argv, 6) // label dx
       && IS_A_FLOAT(argv, 7) // label dy
       && IS_A_FLOAT(argv, 8) // font_style = n, other fstyle flags = 0
       && IS_A_FLOAT(argv, 9) // font size       
       && IS_A_FLOAT(argv, 10) // bcol as 0xXXXXXX int
       && IS_A_FLOAT(argv, 11) // fcol as 0xXXXXXX int
       && IS_A_FLOAT(argv, 12)) // lcol as 0xXXXXXX int
    {
        logd("args accepted\n");
        w = (int) atom_getintarg(0, argc, argv);
        h = (int) atom_getintarg(1, argc, argv);
        iem_inttosymargs(&x->x_gui.x_isa, atom_getintarg(2, argc, argv));
        iemgui_new_getnames(&x->x_gui, 3, argv);
        ldx = (int) atom_getintarg(6, argc, argv);
        ldy = (int) atom_getintarg(7, argc, argv);
        iem_inttofstyle(&x->x_gui.x_fsf, atom_getintarg(8, argc, argv));
        fs = (int) atom_getintarg(9, argc, argv);
        bflcol[0] = (int) atom_getintarg(10, argc, argv);
        bflcol[1] = (int) atom_getintarg(11, argc, argv);
        bflcol[2] = (int) atom_getintarg(12, argc, argv);
        
    } else {
        iemgui_new_getnames(&x->x_gui, 3, 0);
    }

    x->x_gui.x_draw = (t_iemfunptr) ext_draw;
    x->x_gui.x_glist = (t_glist *) canvas_getcurrent();
    
    // snd/rcv
    x->x_gui.x_fsf.x_snd_able = 1;
    x->x_gui.x_fsf.x_rcv_able = 1;
    if (!strcmp(x->x_gui.x_snd->s_name, "empty")) x->x_gui.x_fsf.x_snd_able = 0;
    if (!strcmp(x->x_gui.x_rcv->s_name, "empty")) x->x_gui.x_fsf.x_rcv_able = 0;
    if (x->x_gui.x_fsf.x_rcv_able) pd_bind(&x->x_gui.x_obj.ob_pd, x->x_gui.x_rcv);
    iemgui_verify_snd_ne_rcv(&x->x_gui);    
    // fonts
    if (x->x_gui.x_fsf.x_font_style == 1) strcpy(x->x_gui.x_font, "helvetica");
    else if(x->x_gui.x_fsf.x_font_style == 2) strcpy(x->x_gui.x_font, "times");
    else { x->x_gui.x_fsf.x_font_style = 0; strcpy(x->x_gui.x_font, "courier"); }
    if (fs < 4) fs = 4;
    x->x_gui.x_fontsize = fs;
    x->x_gui.x_w = w;
    x->x_gui.x_h = h;
    // label
    x->x_gui.x_ldx = ldx;
    x->x_gui.x_ldy = ldy;    
    // colors
    iemgui_all_colfromload(&x->x_gui, bflcol);
    
    return (void *) x;
}

void ext_free(t_ext *x) {
    if(x->x_gui.x_fsf.x_rcv_able) {
        pd_unbind(&x->x_gui.x_obj.ob_pd, x->x_gui.x_rcv);
    }
    gfxstub_deleteforkey(x);
}

void ext_setup(void) {
    
    ext_class = class_new(gensym("ext"), (t_newmethod)ext_new, (t_method)ext_free, sizeof(t_ext), 0, A_GIMME, 0);
    
    ext_widgetbehavior.w_getrectfn = ext_getrect;
    ext_widgetbehavior.w_displacefn = iemgui_displace;
    ext_widgetbehavior.w_selectfn = iemgui_select;
    ext_widgetbehavior.w_activatefn = NULL;
    ext_widgetbehavior.w_deletefn = iemgui_delete;
    ext_widgetbehavior.w_visfn = iemgui_vis;
    ext_widgetbehavior.w_clickfn = NULL;

    class_setwidget(ext_class, &ext_widgetbehavior);
}
